/******************************************************************************
*
* microstamp11.c - hardware initialization for "MicroStamp11 GNU-C runtime"
*
* Copyright (c) 2002, 2004, Don Woodward.
*
*
* This file is part of "MicroStamp11 GNU-C runtime".
*
* "MicroStamp11 GNU-C runtime" is free software; you can redistribute
* it and/or modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2 of the
* License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/

#include "microstamp11.h"
#include "hc11ed0.h"

/******************************************************************************
* module microstamp11 static variables
  NOTE NOTE NOTE: Do NOT initialize these variables, we want them to go into the
  bss section, not the data section!
  The data section will be located in RAM in the output file and will overwrite
  the MicroStamp11 eeprom loader as it loads the s records.
  Instead make all static variables uninitialized so that they go into the bss section
  and will be zeroed by gnu at runtime.  You can further initialize these variables
  in InitRuntime();
*/
static u8           s_cSerialPortCharRead = 0;
static boolean      s_bSerialPortCharPending = FALSE;
static TimeStruct   s_kernelTime = { 0, 0 };
static u16          s_nKernelTimeMicroseconds = 0;
static volatile u16 s_nStackMax = 0xFF;                 /* put this variable last - it will be highest
                                                           in RAM and as such will clobber the stack
                                                           most effectively should the stack grow down
                                                           into global variable space */

/******************************************************************************
* module microstamp11 #defines */
#define TIMERMASK           0x40;   /* No timer prescale; RTI enabled */
#define TIMERRATEMASK       0x8;    /* (pactl) 4.096 msec interrupt with 8 MHz xtal */
#define PORTDDATAMASK       0x2;	/* (ddrd) portd data direction: input, except txd */
#define BUAD9600MASK        0x30;	/* 9600 baud on 8Mhz crystal */
#define SCCR2TXRXRXIMASK    0x2C;   /* (sccr2)enable tx & rx and Rx Interrupt */

/******************************************************************************
* function __premain() - called directly by gnu (crt0.o)                     */
int
__premain()
{
    /*  premain() is called by gnu prior to initializing the .data and .bss sections.
        we'll use it to initialize the hardware, but we won't enable interrupts until
        we're inside function main().  That way interrupts won't be enabled until
        gnu has initialized all program global variables.

        A few words on gnu's memory organization:
            1.  gnu has a section called .page0 in which is puts its 'soft' registers.
                This section is placed in the output (elf) file (located at 0x0040 by memory.x),
                but when we make the s-record file with objcopy, we strip the .page0 section.  We do
                this because loading s-records into ram will overwrite the program running
                in chip ram used to load the microstamp 11.  Technically we should
                have code here in __premain() that initializes .page0 to zero, but it
                doesn't seem to matter so we don't bother.
            2.  Immediately following the .page0 section, gnu locates the .data section which consists of
                initialized global variables.  The .data section is not placed in ram in the
                output, rather the initializer values are located in EEPROM (after the .text section)
                and crt0.o contains code that will copy the values from EEPROM into their proper
                locations in RAM.  This code gets called *after* __premain() and before main().
            3.  Immediately following the .data section, gnu locates the .bss section which consists
                of uninitialized global variables.  The .bss section is not placed in the output, rather
                crt0.o contains code that initializes the .bss section to zero.  This code gets called
                immediately after the .data section is initialized.
    */
            
    
    
    /*  Let's initialize the hardware.
        Interrupts are already disabled, and we won't enable them when we're done.      */

    CONFIG = 4;                     /* disable COP (watchdog) */
    TMSK2 = TIMERMASK;              /* enable real-time interrupt(RTI) */
	PACTL = TIMERRATEMASK;          /* init the rti rate */
    DDRD = PORTDDATAMASK;           /* init port d */
    BAUD = BUAD9600MASK;            /* set it to 9600 baud */
    SCCR2 = SCCR2TXRXRXIMASK;       /* enable SCI tx/rx and rx int. */
    s_cSerialPortCharRead = SCDR;   /* dummy read to flush receive buffer */

    return 0;
}

/******************************************************************************
* _microstamp11_check_stack() */
void
_microstamp11_check_stack()
{
    u16 nStackMaxOld = s_nStackMax;
    asm ("sts s_nStackMax");
    if (nStackMaxOld < s_nStackMax) s_nStackMax = nStackMaxOld;
}

/******************************************************************************
* _microstamp11_handle_sci() */
void
_microstamp11_handle_sci()
{
    s_cSerialPortCharRead = SCSR;      /* clear DRF */
    s_cSerialPortCharRead = SCDR;
    s_bSerialPortCharPending = TRUE;
}

/******************************************************************************
* _microstamp11_handle_rti() */
void
_microstamp11_handle_rti()
{
    /* each tick is 4.096 ms */

    s_kernelTime.m_nMilliseconds += 4;
    s_nKernelTimeMicroseconds += 96;

    if (s_nKernelTimeMicroseconds > 999)
    {
        s_nKernelTimeMicroseconds -= 1000;
        s_kernelTime.m_nMilliseconds++;
    }
    
    if (s_kernelTime.m_nMilliseconds > 999)
    {
        s_kernelTime.m_nMilliseconds -= 1000;
        s_kernelTime.m_nSeconds++;
    }
}

/******************************************************************************
* PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
* serialGetChar() */
u8
serialGetChar()
{
    /* blocks until char received */
    u8 c;
    while (s_bSerialPortCharPending == FALSE);
    c = s_cSerialPortCharRead;
    s_bSerialPortCharPending = FALSE;
    return c;
}

/******************************************************************************
* serialPutChar() */
void
serialPutChar(u8 c)
{
    /* puts character to serial port */
    if (c == '\n') c = '\r';
again:
    while (((s8)SCSR) >= 0);
    SCDR = c;
    if (c == '\r') { c = '\n'; goto again; }
}

/******************************************************************************
* serialCharPending() */
boolean
serialCharPending()
{
    /* returns non-zero if char is pending (input) on the serial port */
    return s_bSerialPortCharPending;
}

/******************************************************************************
* serialPutString() */
void
serialPutString(const char * pString)
{
    while (*pString) serialPutChar(*pString++);
}

/******************************************************************************
* serialPutLine() */
void
serialPutLine(const char * pLine)
{
    while (*pLine) serialPutChar(*pLine++);
    serialPutChar('\n');
}

/******************************************************************************
* serialGetLine() */
void
serialGetLine(char * pBuff, u16 nBuffLen, boolean bEcho)
{
    u16 nNumChars = 0;
    u8 ch;

    if ((NULL == pBuff) || (0 == nBuffLen)) return;

    for (;;)
    {
        ch = serialGetChar();

        if ((ch == '\n') || (ch == '\r')) break;
        if (ch == 3)
        {
            if (bEcho) serialPutString("^C");
            nNumChars = 0;
            break;
        }

        if (ch == 8)
        {
            if (0 == nNumChars)
            {
                if (bEcho) serialPutChar(7);
            }
            else
            {
                if (bEcho)
                {
                    serialPutChar(ch);
                    serialPutChar(' ');
                    serialPutChar(ch);
                }
                nNumChars--;
            }
            continue;
        }

        if (nNumChars == (nBuffLen - 1))
        {
            if (bEcho) serialPutChar(7);
        }
        else
        {
            pBuff[nNumChars++] = ch;
            if (bEcho) serialPutChar(ch);
        }
    }

    pBuff[nNumChars] = 0;
    if (bEcho) serialPutChar('\n');
}

/******************************************************************************
* getKernelTime() */
void
getKernelTime(TimeStruct * pTimeStructToSet)
{
    disable();
    *pTimeStructToSet = s_kernelTime;
    enable();
}

/******************************************************************************
* getMaxStack() */
void
getMaxStack(u16 * pMaxStackToSet)
{
    disable();
    *pMaxStackToSet = s_nStackMax;
    enable();
}

/******************************************************************************
* sleep() */
void
sleep(u32 nMilliseconds)
{
    TimeStruct currentTime;
    TimeStruct stopTime;

    getKernelTime(&currentTime);
    stopTime = currentTime;

/*  - Avoid using DIV and MOD, they cause a lot of code to be generated
    stopTime.m_nSeconds += nMilliseconds / 1000;
    stopTime.m_nMilliseconds += nMilliseconds % 1000;
    if (stopTime.m_nMilliseconds > 999)
    {
        stopTime.m_nSeconds++;
        stopTime.m_nMilliseconds -= 1000;
    }
*/
    
    stopTime.m_nMilliseconds += nMilliseconds;
    while (stopTime.m_nMilliseconds > 999)
    {
        stopTime.m_nSeconds++;
        stopTime.m_nMilliseconds -= 1000;
    }

    for (;;)
    {
        getKernelTime(&currentTime);
        if (currentTime.m_nSeconds > stopTime.m_nSeconds) break;
        if ((currentTime.m_nSeconds == stopTime.m_nSeconds) && (currentTime.m_nMilliseconds >= stopTime.m_nMilliseconds)) break;
    }
}

/******************************************************************************
*   LOG
*******************************************************************************
*   03-Jan-02   dwoodward   created
*
*/
